@@ -2,7 +2,7 @@ package Cache::Memcached::Managed;
 
 # Make sure we have version info for this module
 
-$VERSION = '0.20';
+$VERSION = '0.21';
 
 # Make sure we're as strict as possible
 # With as much feedback that we can get
@@ -65,87 +65,93 @@ BEGIN {
 # OUT: 1 instantiated object
 
 sub new {
-
-# Obtain the class
-# Obtain the parameter hash
-
     my $class = shift;
     my %self = @_ < 2 ? (data => (shift || '127.0.0.1:11211')) : @_;
 
-# If we want to force an inactive object
-#  Make sure we have the code to do it
-#  Return the inactive object
-
+    # want to force an inactive object
     if (delete $self{'inactive'}) {
         require Cache::Memcached::Managed::Inactive;
         return Cache::Memcached::Managed::Inactive->new;
     }
 
-# Set the default expiration if not set already
-# Set the delimiter if not set already
-# Set the namespace if not set already
+    # set defaults
+    $self{expiration} = $expiration  if !$self{expiration};
+    $self{delimiter}  = $default_del if !length( $self{delimiter} || '' );
+    $self{namespace}  = $>           if !defined $self{namespace};
 
-    $self{'expiration'} = $expiration  unless $self{'expiration'};
-    $self{'delimiter'}  = $default_del unless length( $self{'delimiter'}||'' );
-    $self{'namespace'}  = $>           unless defined $self{'namespace'};
-
-# Set the group names if not set already
-# Set the group names as hash for easy checking
-
-    $self{'group_names'}  =
-     [$self{'group_names'} ? sort @{$self{'group_names'}} : 'group'];
-    $self{'_group_names'} = {map {$_ => undef} @{$self{'group_names'}}};
+    # set group names
+    $self{group_names}  =
+      [ $self{group_names} ? sort @{ $self{group_names} } : 'group' ];
+    $self{_group_names} = { map { $_ => undef } @{ $self{group_names} } };
 
     # obtain client class
     my $memcached_class = $self{memcached_class} ||= 'Cache::Memcached';
     die $@ if !eval "require $memcached_class; 1";
 
-# For both backends we need
-#  Reloop if there is nothing there
-#  Reloop if we already have blessed object
-#  Make sure we can make Cache::Memcached objects ourselves
-#  Create shortcut to type
-#  Initialize parameters
+    # check both backends
+    my @all_servers;
+  BACKEND:
+    foreach ( qw( data directory ) ) {
+        my $spec = $self{$_};
+        next BACKEND if !$spec;
+        next BACKEND if blessed $spec;
 
-    foreach (qw(data directory)) {
-        next unless $self{$_};
-        next if blessed $self{$_};
-        my $type = reftype $self{$_};
+        # assume a single server spec
         my $parameters;
-
-#  If it is just a scalar
-#   Assume it's a single server specification only
-#  Elsif it is a list ref
-#   Assume it's a multiple server specification only
-#  Elseif it is a hash ref
-#   Assume it is a complete parameter specification
-#  Else
-#   Quit because we don't know how to handle
-#  Attempt to create the object
-
-        if (!defined $type) {
-            $parameters = {servers => [split ',',$self{$_}]};
-        } elsif ($type eq 'ARRAY') {
-            $parameters = {servers => $self{$_}};
-        } elsif ($type eq 'HASH') {
-            $parameters = $self{$_};
-        } else {
-            die "Don't know how to handle '$self{$_}' as server specification";
+        my $type = reftype $spec;
+        if ( !defined $type ) {
+            my @servers = split ',', $spec;
+            push @all_servers, @servers;
+            $parameters = { servers => \@servers };
         }
-        $self{$_} = $memcached_class->new( $parameters );
+
+        # list ref of servers
+        elsif ( $type eq 'ARRAY' ) {
+            push @all_servers, @{$spec};
+            $parameters = { servers => $spec };
+        }
+
+        # ready made parameter hash
+        elsif ( $type eq 'HASH' ) {
+            $parameters = $spec;
+
+            # attempt to find server spec in there
+            $spec = $parameters->{servers};
+            $type = reftype $spec;
+            if ( !defined $type ) {
+                push @all_servers, split ',', $spec;
+            }
+            elsif ( $type eq 'HASH' ) {
+                push @all_servers, @{$spec};
+            }
+
+            # huh?
+            else {
+                undef $parameters;
+            }
+        }
+
+        # huh?
+        die "Don't know how to handle '$spec' as server specification"
+          if !$parameters;
+
+        # create the object for the backend
+        $self{$_} = $memcached_class->new($parameters);
     }
 
-# Quit now if no data server available
-# Set directory server if there was no data server
-# Remember the pid for fork checking
+    # huh?
+    die "No valid data server specification found" if !blessed $self{data};
 
-    die "No valid data server specification found" unless blessed $self{'data'};
-    $self{'directory'} = $self{'data'} unless blessed $self{'directory'};
-    $self{'_last_pid'} = $$;
+    # set directory server as data server if there was no data server
+    $self{directory} = $self{data} if !blessed $self{directory};
 
-# Return the hash as blessed object
+    # remember the pid for fork checking
+    $self{_last_pid} = $$;
 
-    bless \%self,$class;
+    # set server specification
+    $self{servers} = [ sort @all_servers ];
+
+    return bless \%self,$class;
 } #new
 
 #---------------------------------------------------------------------------
@@ -389,7 +395,7 @@ sub flush_all {
 
     my ($self,$interval) = @_;
     my $data = $self->data;
-    my @server = @{$data->{'servers'}};
+    my @server = $self->servers;
 
 # Use default interval if none specified
 # Initialize number of servers flushed
@@ -749,27 +755,28 @@ sub replace { shift->_do( 'replace',@_ ) } #replace
 # OUT: 1 returns true
 
 sub reset {
-
-# Obtain the object
-# Obtain local copy of data and directory object
-
     my $self = shift;
-    my ($data,$directory) = ($self->data,$self->directory);
 
-# For all of the Cache::Memcached objects we need to handle
-#  Disconnect all sockets
-#  Kickstart connection logic
+    # obtain local copy of data and directory object
+    my ( $data, $directory ) = ( $self->data, $self->directory );
 
-    foreach ($data == $directory ? ($data) : ($data,$directory)) {
-        $_->disconnect_all;
-        $_->forget_dead_hosts; # not sure official C::Memcached API
+    # all of the Cache::Memcached objects we need to handle
+    foreach ( $data == $directory ? ($data) : ( $data, $directory ) ) {
+
+        # disconnect all sockets
+        $_->disconnect_all    if $_->can('disconnect_all');;
+
+        # kickstart connection logic
+        $_->forget_dead_hosts if $_->can('forget_dead_hosts');
     }
 
-# Make sure we try to connect again
-# Set last pid used flag
-
+    # make sure we try to connect again
     $self->_mark_connected;
+
+    # set last pid used flag
     $self->{'_last_pid'} = $$;
+
+    return 1;
 } #reset
 
 #---------------------------------------------------------------------------
@@ -805,14 +812,9 @@ sub set { shift->_do( 'set',@_ ) } #set
 
 sub servers {
 
-# Obtain the object
-# Obtain the server specifications, weed out the double ones
-# Return the specifications sorted or as a hash ref
-
-    my $self = shift;
-    my %server =
-     map {$_ => undef} map {@{$self->$_->{'servers'}}} qw(data directory);
-    return wantarray ? sort keys %server : \%server;
+    return wantarray
+      ? @{ shift->{servers} }
+      : { map { $_ => undef } @{ shift->{servers} } };
 } #servers
 
 #---------------------------------------------------------------------------
@@ -881,7 +883,7 @@ sub stats {
     my %todo = @_ ? map {$_ => undef} @_ : %{$self->servers};
     my %result;
     foreach my $cache ($self->data,$self->directory) {
-        foreach my $host (@{$cache->{'servers'}}) {
+        foreach my $host ( $self->servers ) {
             next unless exists $todo{$host} and not exists $result{$host};
             $result{$host} = {
              map {s#^STAT ##; split m#\s+#}
@@ -1269,22 +1271,26 @@ sub _do {
         return $result;
     }
 
-# Obtain the directory server
-# If we still don't have a good result
-#  If we can obtain a bucket for this data key
-#   If we can lose the prefix
-#    Increment error for given server
-#  Block all access for this process
-#  Return indicating error
-
+    # still don't have a good result
     my $directory = $self->directory;
-    unless ($result) {
-        if (my $bucket = $data->get_sock( $data_key )) {
-            if ($bucket =~ s#^Sock_##) {
-                $directory->add( $bucket,1 ) unless $directory->incr( $bucket );
+    if ( !$result ) {
+
+        # can get the bucket
+        if ( $data->can('get_sock') ) {
+            if ( my $bucket = $data->get_sock($data_key) ) {
+
+                # can lose prefix, increment error on server
+                if ( $bucket =~ s#^Sock_## ) {
+                    $directory->add( $bucket, 1 )
+                      if !$directory->incr($bucket);
+                }
             }
         }
+
+        # block all access for this process
         $self->{'_disconnected'} = 1;
+
+        # return indicating error
         return undef;
     }
 
@@ -1491,18 +1497,17 @@ sub _mark_disconnected {
 # OUT: 1..N response lines
 
 sub _morelines {
+    my ( $self, $cache, $host, $send, $bucket ) = @_;
 
-# Obtain the parameters
-# Return now if we couldn't get a socket for the indicated bucket
+    # don't have any sock to host mapping, so quit
+    return if !$cache->can('sock_to_host');
 
-    my ($self,$cache,$host,$send,$bucket) = @_;
-    return unless my $socket = $cache->sock_to_host( $host );
+    # couldn't get a socket for given host
+    return unless my $socket = $cache->sock_to_host($host);
 
-# Return result of fetch, without newlines
-
-    map {
+    return map {
      s#[\r\n]+$##; m#^(?:END|ERROR)# ? () : ($_)
-    } $cache->run_command( $socket,$send."\r\n" );
+    } $cache->run_command( $socket, $send. "\r\n" );
 } #_morelines
 
 #---------------------------------------------------------------------------
@@ -1516,27 +1521,29 @@ sub _morelines {
 # OUT: 1 response or whether expected response returned
 
 sub _oneline {
+    my ( $self, $cache, $send, $bucket, $expect ) = @_;
 
-# Obtain the parameters
-# Return now if we couldn't get a socket for the indicated bucket
+    # can't get any socket, so quit
+    return if !$cache->can('get_sock');
 
-    my ($self,$cache,$send,$bucket,$expect) = @_;
+    # couldn't get a socket for the indicated bucket
     return unless my $socket = $cache->get_sock( [$bucket || 0,0] );
 
-# Make sure we can call a "_oneline" compatible method
-
-    $_oneline ||= $cache->can( '_oneline' ) || $cache->can( '_write_and_read' )
-      or die "Unsupported version of ".(blessed $cache)."\n";
-
-# Send the request and obtain the response
-# Return response if we don't want to check here
-# Return whether result was expected
+    # make sure we can call a "_oneline" compatible method
+    $_oneline ||=
+      $cache->can( '_oneline' ) ||
+      $cache->can( '_write_and_read' )
+      or die "Unsupported version of " . ( blessed $cache ) . "\n";
 
+    # obtain response
     my $response = defined $send
-     ? $_oneline->( $cache, $socket, $send."\r\n" )
+     ? $_oneline->( $cache, $socket, $send . "\r\n" )
      : $_oneline->( $cache, $socket );
-    return $response unless defined $expect;
-    $response and $expect."\r\n" eq $response;
+
+    # nothing to check against, just give back what we got
+    return $response if !defined $expect;
+
+    return ( $response and $expect."\r\n" eq $response );
 } #_oneline
 
 #---------------------------------------------------------------------------
@@ -1554,6 +1561,33 @@ sub _unique_key {
     join $_[0]->delimiter,$server,$$,time,++$unique;
 } #_unique_key
 
+#---------------------------------------------------------------------------
+# _spec2servers
+#
+# Converts server spec to list ref of servers
+#
+#  IN: 1 server spec
+#      2 recursing flag (only used internally)
+# OUT: 1 list ref of servers
+
+sub _spec2servers {
+    my ( $spec, $recursing ) = @_;
+
+    # assume scalar definition if not a ref
+    my $type = reftype $spec;
+    if ( !defined $type ) {
+        return [ split ',', $spec ];
+    }
+
+    # list ref of servers
+    elsif ( $type eq 'ARRAY' ) {
+        return $spec;
+    }
+
+    # huh?
+    die "Don't know how to handle '$spec' as server specification";
+}    #_spec2servers
+
 #---------------------------------------------------------------------------
 
 __END__
@@ -1586,7 +1620,7 @@ Cache::Memcached::Managed - provide API for managing cached information
 
 =head1 VERSION
 
-This documentation describes version 0.20.
+This documentation describes version 0.21.
 
 =head1 DIFFERENCES FROM THE Cache::Memcached API
 
@@ -1616,7 +1650,7 @@ a Class::DBI object).  If necessary, a unique ID can be created automatically
 
 The caller's package provides an identifying version (by default), allowing
 differently formatted data-structures caused by source code changes, to live
-seperately from each other in the cache.
+separately from each other in the cache.
 
 =head2 namespace support
 
@@ -1746,7 +1780,7 @@ The data server can be obtained with the L<data> object.
 
 The data key identifies a piece of data in the L<"data server">.  It is
 formed by appending the namespace (by default the user id of the process),
-L<version>, L<key> and L<ID>, seperated by the L<delimiter>.
+L<version>, L<key> and L<ID>, separated by the L<delimiter>.
 
 If a scalar value is specified as an ID, then that value is used.
 
@@ -1807,7 +1841,7 @@ L<"group management">.  If no L<directory> server was specified, then the
 data server will be assumed.
 
 If there are multiple memcached servers used for the L<"data server">, then
-it is advised to use a seperate directory server (as a failure in one of
+it is advised to use a separate directory server (as a failure in one of
 the memcached backend servers will leave you with an incomplete directory
 otherwise).
 
@@ -1883,7 +1917,7 @@ following fields:
 The specification of the memcached server backend(s) for the L<"data server">.
 It should either be:
 
- - string with comma seperated memcached server specification
+ - string with comma separated memcached server specification
  - list ref with memcached server specification
  - hash ref with Cache::Memcached object specification
  - blessed object adhering to the Cache::Memcached API
@@ -1921,7 +1955,7 @@ some encoding issues within L<Cache::Memcached> regarding null bytes.
 The specification of the memcached server backend(s) for the
 L<"directory server">.  It should either be:
 
- - string with comma seperated memcached server specification
+ - string with comma separated memcached server specification
  - list ref with memcached server specification
  - hash ref with Cache::Memcached object specification
  - blessed object adhering to the Cache::Memcached API
@@ -2620,7 +2654,7 @@ This would put the value C<$value> into the cache, linked to the group
 'event1'.  Since we're not interested in the id of the event, but want to
 make sure it is always unique, the pseudo id ':unique' is specified.
 
-A recurring proces, usually a cron job, would then need to do the following
+A recurring process, usually a cron job, would then need to do the following
 to grab all of the values cached:
 
  my @value = $cache->grab_group( group => 'event1' );
@@ -2734,7 +2768,7 @@ test-suite).
 =head1 THEORY OF OPERATION
 
 The group management is implemented by keeping a type of directory information
-in a (seperate) directory memcached server.
+in a (separate) directory memcached server.
 
 For each L<group|"group management"> one directory key is maintained in the
 directory memcached server.  This key consists of the string
